#!/usr/bin/python3
# coding: utf-8
"""
Add the workflow to a trace

Currently, it is hard coded to a single trace fileº

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2023
"""

import pyswf.pyswf as pyswf
# import pyswf
import argparse  # to handle command line options
import os  # to create and manage folders


def add_horizontal_wf(swf_trace, num_jobs, num_wrapped: int, submit_time, target_uid, runtime=1800, alloc_cpus=96,
                      requested_time=2400, partition_id=1):
    """
    Add a horizontal workflow
    :param swf_trace: trace that the workflow will be added to
    :param num_jobs: number of jobs to add
    :param num_wrapped: number of jobs per wrapper
    :param submit_time: moment in which the workflow will be submitted
    :param target_uid: the uid of the user that will launch the workflow
    :param runtime: the runtime for the jobs (it is assumed the same for all of them)
    :param alloc_cpus: number of allocated cpus per job (assumed the same for all of them)
    :param requested_time: wallclock of each of the jobs (assumed the same for all of them)
    :param partition_id: partition in which it will be launched
    :return: None with swf_trace changed inplace
    """
    to_be_wrapped = num_jobs

    while to_be_wrapped - num_wrapped >= 0:
        # we assume that the gid is the same as the uid
        swf_trace.add_job(submit_time, num_wrapped * runtime, alloc_cpus, target_uid, requested_time, gid=target_uid,
                          partition_id=partition_id)
        to_be_wrapped -= num_wrapped

    # if there are still jobs to add, we add them individually. This mimics Autosubmit's FLEXIBLE policy with
    # MIN_JOBS = num_wrapped
    if to_be_wrapped > 0:
        for i in range(to_be_wrapped):
            swf_trace.add_job(submit_time, runtime, alloc_cpus, target_uid, requested_time, gid=target_uid,
                              partition_id=partition_id)

def add_vertical_wf(swf_trace, num_jobs: int, num_wrapped: int, submit_time: int, target_uid: int,
                    runtime: int = 1800, alloc_cpus: int = 96, requested_time: int = 2400, partition_id: int = 1):
    """
    Add a vertical workflow
    :param swf_trace: trace that the workflow will be added to
    :param num_jobs: number of jobs to add
    :param num_wrapped: number of jobs per wrapper
    :param submit_time: moment in which the workflow will be submitted
    :param target_uid: the uid of the user that will launch the workflow
    :param runtime: the runtime for the jobs (it is assumed the same for all of them)
    :param alloc_cpus: number of allocated cpus per job (assumed the same for all of them)
    :param requested_time: wallclock of each of the jobs (assumed the same for all of them)
    :param partition_id: partition in which it will be launched
    :return: None with swf_trace changed inplace
    """
    to_be_wrapped = num_jobs

    while to_be_wrapped - num_wrapped >= 0:
        # we assume that the gid is the same as the uid
        swf_trace.add_job(submit_time, num_wrapped*runtime, alloc_cpus, target_uid, requested_time, gid=target_uid,
                          partition_id=partition_id)
        to_be_wrapped -= num_wrapped

    # if there are still jobs to add, we add them individually. This mimics Autosubmit's FLEXIBLE policy with
    # MIN_JOBS = num_wrapped
    if to_be_wrapped > 0:
        for i in range(to_be_wrapped):
            # handle dependencies ...
            preceding_job = swf_trace.job_list[swf_trace.job_list.uid == target_uid].iloc[-1]
            preceding_job_id = preceding_job.job_id
            preceding_job_submit = preceding_job.submit_time
            preceding_job_runtime = preceding_job.runtime
            swf_trace.add_job(preceding_job_submit + preceding_job_runtime, runtime, alloc_cpus,
                              target_uid, requested_time, gid=target_uid, partition_id=partition_id,
                              preceding=preceding_job_id)

if __name__ == "__main__":

    # PARSE STUFF
    parser = argparse.ArgumentParser(description='Add a workflow to a swf file. This script will generate two swf files:'
                                                 ' one for unwrapped, another for the wrapped')
    parser.add_argument('input', type=str, help='path where the original trace is located')
    parser.add_argument('output', type=str, help='path where the trace with workflow will be stored')
    parser.add_argument('jobs', type=int, help='number of jobs in the workflow')
    parser.add_argument('runtime', type=int, help='runtime of the jobs')
    parser.add_argument('cores', type=int, help='cores required by each of the jobs')
    parser.add_argument('user', type=int, help='user that will launch the job')
    parser.add_argument('group', type=int, help='group of the user that will launch the jobs')
    parser.add_argument('type', type=str, choices=('vertical', 'horizontal'), help="Type of workflow that will "
                                                                                   "be generated. It can either be "
                                                                                   "vertical or horizontal.")
    parser.add_argument('wrapped', type=int, help="Number of jobs per wrapper")
    parser.add_argument('workflow', type=str, help="Name of the workflow")
    parser.add_argument('platform', type=str, help="Name of the target hpc platform") 
    parser.add_argument('--wallclock', '-w', type=int, help="wallclock of the job")
    parser.add_argument('--fairshare', '-f', type=float, help="Fairshare value that the target user will have") 
    parser.add_argument('--no-wf-run', '-n', type=str, help="Path of the run without wf to compute fairshare ")
    parser.add_argument('--partition', '-p', type=int, help="Partition id of the target job") 

    args = parser.parse_args()

    if args.fairshare and (args.no_wf_run is None):
        raise argparse.ArgumentError("If fairshare is set, it is necessary to provide a path to a run without workflow.")

    # if we wrap more jobs that there are available
    if args.wrapped > args.jobs:
        wrapped = args.jobs  
    else:
        wrapped = args.wrapped

    if args.wallclock is None:
        wallclock = args.runtime

    if args.partition is None:
        partition = 1

    if args.fairshare is None:
        fairshare = 1
        fairshare_runtime = 0

    workload = pyswf.SWFtrace.from_swf(args.input)

    workload.note += f"\n;Note: {args.type} workflow added to the user {args.user}"

    # trace without workflow used for computing fairshare
    # TODO: under static workloads is this necessary?
    if args.no_wf_run is not None:
        no_wf = pyswf.SWFtrace.from_slurm_trace(args.no_wf_run)

    # path where we store the traces
    store_path = args.output 

    if not os.path.isdir(store_path):
        print(f"[static_add_wf.py] creating folder {store_path}")
        os.mkdir(store_path)

    # copy original job list since we will be changing it in every function
    clean_job_list = workload.job_list.copy()

    if args.type == "vertical":
        if not os.path.isdir(f"{store_path}/{str(fairshare)}"):
            print(f"[static_add_wf.py] creating folder {store_path}/{str(fairshare)}")
            os.mkdir(f"{store_path}/{str(fairshare)}")

        # vertical unwrapped
        add_vertical_wf(workload, args.jobs, 1, 0, args.user, args.runtime, args.cores, wallclock, partition)
        workload.to_swf(f"{store_path}/{str(fairshare)}/{args.platform}_{args.workflow}_{args.type}_static.swf")
        workload.job_list = clean_job_list.copy()

        # wrapped vertical 
        add_vertical_wf(workload, args.jobs, 7, 0, args.user, args.runtime, args.cores, wallclock, partition)
        workload.to_swf(f"{store_path}/{str(fairshare)}/{args.platform}_{args.workflow}_wrapped_{args.type}_static.swf")
        workload.job_list = clean_job_list
        
    # then we have a horizontal workflow
    else:
        if not os.path.isdir(f"{store_path}/{str(fairshare)}"):
            print(f"[static_add_wf.py] creating folder {store_path}/{str(fairshare)}")
            os.mkdir(f"{store_path}/{str(fairshare)}")

        # horizontal unwrapped
        add_horizontal_wf(workload, args.jobs, 1, 0, args.user, args.runtime, args.cores, wallclock, partition)
        workload.to_swf(f"{store_path}/{str(fairshare)}/{args.platform}_{args.workflow}_{args.type}_static.swf")
        workload.job_list = clean_job_list

        # wrapped horizontal 
        add_horizontal_wf(workload, args.jobs, 12, 0, args.user, args.runtime, args.cores, wallclock, partition)
        workload.to_swf(f"{store_path}/{str(fairshare)}/{args.platform}_{args.workflow}_wrapped_{args.type}_static.swf")
        workload.job_list = clean_job_list
